/* Regular expression search & replace */

#include <utils/p2c.h>


#define REGEX_G
#include <utils/regex.h>
#include <utils/strings.h>


#define checking        false

#define quote           1
#define bunch           2
#define close_bunch     3
#define invert          4
#define through         5
#define kleene          6
#define occur           7
#define some            8
#define eol             9
#define all             10
#define alternative     11
#define qnot            12
#define left_paren      13
#define right_paren     14
#define count           15
#define close_count     16
#define continue_       17
#define star            18
#define begword         19
#define endword         20
#define charsmax        20

#define end_            "\022"
#define quote_          "\001"
#define letter_         "\002"
#define bunch_          "\003"
#define all_            "\004"
#define eol_            "\005"
#define kleene_         "\006"
#define occur_          "\007"
#define some_           "\b"
#define count_          "\t"
#define not_            "\n"
#define alternative_    "\013"
#define lparen_         "\f"
#define rparen_         "\015"
#define nparen_         "\016"
#define begword_        "\017"
#define endword_        "\020"
#define ci_             "\021"


/* Local variables for compile: */
struct LOC_compile {
	char *ex, *rex;
	char chars[256];
	int cimode;
	size_t i;
	long parcnt;
} ;

static void expr (int parenflag, struct LOC_compile *link);

static char *parsebunch(char *rval, struct LOC_compile *link)
{
	uchar a[32];
	long j;
	char ch;
	int inverted, flag;
	char s[256];
	long FORLIM;
	int TEMP;

	inverted = (link->ex[link->i - 1] == link->chars[invert - 1] && link->ex[link->i - 1] != ' ');
	if (inverted)
		link->i++;

	for (j = 0; j <= 255; j++)
	{
		P_clrbits_B(a, j, 0, 3);
		P_putbits_UB(a, j, inverted, 0, 3);
	}

	while (link->i <= strlen(link->ex) && link->ex[link->i - 1] != link->chars[close_bunch - 1])
	{
		ch = link->ex[link->i - 1];
		link->i++;
    	if (link->ex[link->i - 1] == link->chars[through - 1] && link->ex[link->i - 1] != ' ' && link->ex[link->i] >= ch)
		{
			FORLIM = link->ex[link->i];
			for (j = ch; j <= FORLIM; j++)
			{
				P_clrbits_B(a, j, 0, 3);
				P_putbits_UB(a, j, (int)(!inverted), 0, 3);
    		}
    		link->i += 2;
		}
		else
		{
			P_clrbits_B(a, ch, 0, 3);
			P_putbits_UB(a, ch, (int)(!inverted), 0, 3);
		}
	}
	if (link->cimode) {
		for (j = 'a'; j <= 'z'; j++) {
			TEMP = (P_getbits_UB(a, j - 32, 0, 3) || P_getbits_UB(a, j, 0, 3));
			P_clrbits_B(a, j - 32, 0, 3);
			P_putbits_UB(a, j - 32, TEMP, 0, 3);
			P_clrbits_B(a, j, 0, 3);
		}
	}
	if (link->i <= strlen(link->ex))
		link->i++;
	flag = false;
	*s = '\0';
	for (j = 0; j <= 255; j++) {
		if (P_getbits_UB(a, j, 0, 3) != flag) {
			if (flag)
				sprintf(s + strlen(s), "%c", (char)(j - 1));
			else
				sprintf(s + strlen(s), "%c", (char)j);
			flag = P_getbits_UB(a, j, 0, 3);
		}
	}

	if (flag)
		strcat(s, "\377");
	sprintf(rval, "%c%s", (char)(strlen(s) / 2), s);
	return rval;
}

static char *parsecount(char *rval, struct LOC_compile *link)
{
	long min = 0;
	long max = 255;
	char STR1[2];
	char *STR2;

	sprintf(STR1, "%c", link->chars[close_count - 1]);
	
	if (strpos2(link->ex, STR1, link->i) > 0)
	{
		if (isdigit(link->ex[link->i - 1]))
		{
			min = strtol(link->ex + link->i - 1, &STR2, 10);
			link->i = STR2 - link->ex + 1;
		}
		
		if (link->ex[link->i - 1] == link->chars[continue_ - 1] && link->chars[continue_ - 1] != ' ')
		{
			link->i++;
			if (isdigit(link->ex[link->i - 1]))
			{
				max = strtol(link->ex + link->i - 1, &STR2, 10);
				link->i = STR2 - link->ex + 1;
			}
		}
		else
		{
			max = min;
		}
		
		if (link->ex[link->i - 1] == link->chars[close_count - 1])
			link->i++;
	}
	
	/* clamp min and max */
	if (max > 255)
		max = 255;
	
	if (min > max)
		min = max;
	
	sprintf(rval, "%s%c%c", count_, (char)min, (char)max);
	return rval;
}

/* Local variables for expr: */
struct LOC_expr {
	struct LOC_compile *link;
} ;

/* Local variables for term: */
struct LOC_term {
	struct LOC_expr *link;
} ;

static void factor(int parenflag, struct LOC_term *link)
{
	long save;
	char ch;
	char STR1[257];

	struct LOC_compile *p_link = link->link->link;

	if (p_link->i > strlen(p_link->ex))
		return;

	ch = p_link->ex[p_link->i - 1];
	p_link->i++;
	
	if (ch == ' ')
	{
		sprintf(p_link->rex + strlen(p_link->rex), "%c", ch);
		return;
	}

	if (ch == p_link->chars[quote - 1] && ch != ' ' && p_link->i <= strlen(p_link->ex))
	{
		if (isalpha(p_link->ex[p_link->i - 1]))
		{
			strcat(p_link->rex, "\002");
			sprintf(p_link->rex + strlen(p_link->rex), "%c", toupper(p_link->ex[p_link->i - 1]));
			p_link->i++;
			return;
		}
		if (isdigit(p_link->ex[p_link->i - 1]))
		{
			strcat(p_link->rex, "\016");
			if (p_link->ex[p_link->i - 1] == '0')
				strcat(p_link->rex, "\n");
			else
				sprintf(p_link->rex + strlen(p_link->rex), "%c", p_link->ex[p_link->i - 1] - 48);
			p_link->i++;
			return;
		}
		if (p_link->ex[p_link->i - 1] < ' ')
			strcat(p_link->rex, "\001");
		sprintf(p_link->rex + strlen(p_link->rex), "%c", p_link->ex[p_link->i - 1]);
		p_link->i++;
		return;
	}
	if (ch == p_link->chars[bunch - 1] && p_link->i <= strlen(p_link->ex) && p_link->chars[close_bunch - 1] != ' ') {
		strcat(p_link->rex, "\003");
		strcat(p_link->rex, parsebunch(STR1, p_link));
		return;
	}
	if (ch == p_link->chars[all - 1]) {
		strcat(p_link->rex, "\004");
		return;
	}
	if (ch == p_link->chars[eol - 1]) {
		strcat(p_link->rex, "\005");
		return;
	}
	if (ch == p_link->chars[begword - 1]) {
		strcat(p_link->rex, "\017");
		return;
	}
	if (ch == p_link->chars[endword - 1]) {
		strcat(p_link->rex, "\020");
		return;
	}
	if (ch == p_link->chars[left_paren - 1] && p_link->chars[right_paren - 1] != ' ')
	{
		p_link->parcnt++;
		if (p_link->parcnt > 10)
		{
			expr(true, p_link);
		}
		else
		{
			save = p_link->parcnt;
			strcat(p_link->rex, "\f");
			sprintf(p_link->rex + strlen(p_link->rex), "%c", (char)save);
			expr(true, p_link);
			strcat(p_link->rex, "\015");
			sprintf(p_link->rex + strlen(p_link->rex), "%c", (char)save);
		}
		if (p_link->i <= strlen(p_link->ex) && p_link->ex[p_link->i - 1] == p_link->chars[right_paren - 1])
			p_link->i++;
		return;
	}
	if (ch == p_link->chars[kleene - 1] || ch == p_link->chars[star - 1])
	{
		sprintf(p_link->rex + strlen(p_link->rex), "%s%s%s", kleene_, all_, end_);
		return;
	}
	if (ch < ' ')
		strcat(p_link->rex, "\001");
	sprintf(p_link->rex + strlen(p_link->rex), "%c", ch);
}  /*factor*/

static void term(int parenflag, struct LOC_expr *link)
{
	long save;
	char ch;
	int notflag;
	char STR1[256];
	char STR2[256];
	struct LOC_term term;
	term.link = link;
	struct LOC_compile *p_link = link->link;

	while (p_link->i <= strlen(p_link->ex) && (p_link->ex[p_link->i - 1] != p_link->chars[alternative - 1] ||
		p_link->ex[p_link->i - 1] == ' ') && (p_link->ex[p_link->i - 1] != p_link->chars[right_paren - 1] || !parenflag))
	{
		notflag = false;
		while (p_link->i <= strlen(p_link->ex) &&
			p_link->ex[p_link->i - 1] == p_link->chars[qnot - 1] &&
			p_link->ex[p_link->i - 1] != ' ')
		{
			notflag = !notflag;
			p_link->i++;
		}
		if (p_link->i > strlen(p_link->ex))
		{
			notflag = !notflag;
			p_link->i--;
		}
		if (notflag)
			strcat(p_link->rex, "\n");
		save = strlen(p_link->rex) + 1;
		factor(parenflag, &term);
		while (p_link->i <= strlen(p_link->ex) && p_link->ex[p_link->i - 1] != ' ' &&
			(p_link->ex[p_link->i - 1] == p_link->chars[kleene - 1] ||
			p_link->ex[p_link->i - 1] == p_link->chars[occur - 1] ||
			p_link->ex[p_link->i - 1] == p_link->chars[some - 1] ||
			p_link->ex[p_link->i - 1] == p_link->chars[count - 1]))
		{
			ch = p_link->ex[p_link->i - 1];
			p_link->i++;
			if (ch == p_link->chars[kleene - 1]) {
				sprintf(STR1, "%s%s", kleene_, p_link->rex + save - 1);
				strcpy(p_link->rex + save - 1, STR1);
			} else if (ch == p_link->chars[occur - 1]) {
				sprintf(STR1, "%s%s", occur_, p_link->rex + save - 1);
				strcpy(p_link->rex + save - 1, STR1);
			} else if (ch == p_link->chars[some - 1]) {
				sprintf(STR1, "%s%s", some_, p_link->rex + save - 1);
				strcpy(p_link->rex + save - 1, STR1);
			} else {
				sprintf(STR1, "%s%s",
				parsecount(STR2, p_link), p_link->rex + save - 1);
				strcpy(p_link->rex + save - 1, STR1);
			}
			strcat(p_link->rex, "\022");
		}
		if (notflag)
			strcat(p_link->rex, "\022");
	}
  /*count*/
}  /*term*/

static void expr(int parenflag, struct LOC_compile *link)
{
	struct LOC_expr expr;
	long save;
	char STR1[256];

	expr.link = link;
	save = strlen(link->rex) + 1;
	term(parenflag, &expr);
	if (link->i > strlen(link->ex) || link->ex[link->i - 1] != link->chars[alternative - 1] || link->ex[link->i - 1] == ' ')
		return;
	sprintf(STR1, "%s%s", alternative_, link->rex + save - 1);
	strcpy(link->rex + save - 1, STR1);
	strcat(link->rex, "\022");
	
	while (link->i <= strlen(link->ex) && link->ex[link->i - 1] == link->chars[alternative - 1])
	{
		link->i++;
		strcat(link->rex, "\013");
		term(parenflag, &expr);
		strcat(link->rex, "\022");
	}
}  /*expr*/

static void compile(char *ex_, char *rex_, char *chars_, int cimode_)
{
	struct LOC_compile compile;

	compile .ex = ex_;
	compile.rex = rex_;
	strcpy(compile.chars, chars_);
	compile.cimode = cimode_;
	if (*(compile.chars) == '\0')
		strcpy(compile.chars, re_defchars);
	
	while (strlen(compile.chars) < charsmax)
		strcat(compile.chars, " ");
	compile.parcnt = 0;
	if (compile.cimode)
		strcpy(compile.rex, ci_);
	else
		*compile.rex = '\0';
	compile.i = 1;
	expr(false, &compile);
	strcat(compile.rex, end_);
}

void re_compile(char *ex_, char *rex, char *chars)
{
	char ex[256];

	strcpy(ex, ex_);
	compile(ex, rex, chars, false);
}

void re_cicompile(char *ex_, char *rex, char *chars)
{
	char ex[256];

	strcpy(ex, ex_);
	compile(ex, rex, chars, true);
}

static char *charconv(char *rval, char ch)
{
	if (ch > ' ' && (ch & (~127)) == 0)
	{
		sprintf(rval, "%c", ch);
		return rval;
	}
	else if (ch == ' ')
	{
		return strcpy(rval, "sp");
	}
	else
	{
		sprintf(rval, "#%d", ch);
		return rval;
	}
}

void re_dump(char *rex)
{
	size_t i;
	size_t n;
	char STR2[256], STR3[256];

	if (*rex == '\0')
		printf("<null>");

	i = 1;
	while (i <= strlen(rex))
	{
		if (i != 1)
			putchar(' ');

		if (rex[i - 1] > ' ')
		{
			putchar(rex[i - 1]);
		}
		else if (rex[i - 1] == ' ')
		{
			printf("sp");
		}
		else
		{
			putchar('<');
			switch (rex[i - 1])
			{
				case '\022':
					printf("end");
					break;

				case '\001':
					i++;
					printf("quote %s", charconv(STR2, rex[i - 1]));
					break;

				case '\002':
					printf("letter");
					break;

				case '\003':
					printf("bunch");
					i++;
					n = rex[i - 1];
					while (n > 0)
					{
						printf(" %s-%s", charconv(STR2, rex[i]), charconv(STR3, rex[i + 1]));
						i += 2;
						n--;
					}
					break;

				case '\004':
					printf("all");
					break;

				case '\005':
					printf("eol");
					break;

				case '\017':
					printf("begword");
					break;

				case '\020':
					printf("endword");
					break;
				
				case '\006':
					printf("kleene");
					break;

				case '\007':
					printf("occur");
					break;

				case '\b':
					printf("some");
					break;

				case '\t':
					printf("count %d:%d", rex[i], rex[i + 1]);
					i += 2;
					break;

				case '\n':
					printf("not");
					break;

				case '\013':
					printf("alt");
					break;

				case '\f':
				case '\015':
				case '\016':
					switch (rex[i - 1])
					{
						case '\f':
							printf("lparen ");
							break;

						case '\015':
							printf("rparen ");
							break;

						case '\016':
							printf("nparen ");
							break;
					}
					printf("%d", rex[i]);
					i++;
					break;

				case '\021':
					printf("ci");
					break;
			}
			putchar('>');
		}
		i++;
	}
}

/* Local variables for compare: */
struct LOC_compare {
  char *s, *rex, *rpl;
  size_t *p;
  size_t i, savep, countleft, countright;
  long par[2][10];
  int parflag, cimode;
  char matchchar;
} ;

static int comp (struct LOC_compare *link);

static void skiprex(char *rex, struct LOC_compare *link)
{
	char STR1[256];

	while (strcmp((sprintf(STR1, "%c", rex[link->i - 1]), STR1), end_)) {
		switch (rex[link->i - 1]) {

			case '\006':
			case '\007':
			case '\b':
			case '\n':
			case '\013':
				link->i++;
				skiprex(rex, link);
				break;

			case '\t':
				link->i += 3;
				skiprex(rex, link);
				break;

			case '\001':
			case '\f':
			case '\015':
			case '\016':
				link->i += 2;
				break;

			case '\003':
				link->i += rex[link->i] * 2 + 2;
				break;

			default:
				link->i++;
				break;
		}
	}
	link->i++;
}

/* Local variables for comp: */
struct LOC_comp {
	struct LOC_compare *link;
	int wrong;
	long tempright;
} ;

/* Local variables for count__: */
struct LOC_count__ {
	struct LOC_comp *link;
	long min, max, i1;
	int flag;
} ;

static int subcount(long num, struct LOC_count__ *link)
{
	long savep, savei;

	savep = *link->link->link->p;
	savei = link->link->link->i;
	link->link->link->i = link->i1;
	
	if (num < link->max && comp(link->link->link))
	{
		link->link->link->i++;
		if (subcount(num + 1, link))
		{
			link->link->tempright = *link->link->link->p;
			link->flag = (num + 1 < link->min || !comp(link->link->link));
		}
		else
		{
			link->flag = false;
		}
	}
	else
	{
		link->flag = true;
	}
	if (link->flag)
	{
		*link->link->link->p = savep;
		link->link->link->i = savei;
	}

	return link->flag;
}  /*subcount*/

static void count__(long min_, long max_, struct LOC_comp *link)
{
	struct LOC_count__ V;

	V.link = link;
	V.min = min_;
	V.max = max_;
	link->link->countleft = *link->link->p;
	link->tempright = -1;
	V.i1 = link->link->i;
	link->wrong = subcount(0L, &V);
	
	if (link->wrong && V.min == 0) {
		skiprex(link->link->rex, link->link);
		link->tempright = *link->link->p;
		link->wrong = !comp(link->link);
	}

	if (!link->wrong)
		link->link->countright = link->tempright;
}  /*count*/

static int comp(struct LOC_compare *link)
{
	struct LOC_comp V;
	int flag;
	char ch, ch1;
	size_t j, n, savei, savep;
	char STR1[256], STR2[256];
	char STR3[256], STR4[256];

	V.link = link;
	/* write('Comp "', strsub(s,p,999), '" against "'); re_dump(strsub(rex,i,999)); writeln('"'); {****/
	V.wrong = false;
	savei = link->i;
	savep = *link->p;
	while (strcmp((sprintf(STR1, "%c", link->rex[link->i - 1]), STR1), end_) && !V.wrong)
	{
		ch = link->rex[link->i - 1];
		link->i++;
		if (ch >= ' ')
		{
			if (*link->p > strlen(link->s))
			{
				V.wrong = true;
				break;
			}
			if (link->cimode)
				V.wrong = (toupper(link->s[*link->p - 1]) != toupper(ch));
			else
				V.wrong = (link->s[*link->p - 1] != ch);
			(*link->p)++;
			continue;
		}
		
		switch (ch)
		{
			case '\001':
				ch = link->rex[link->i - 1];
				link->i++;
				if (*link->p > strlen(link->s))
				{
					V.wrong = true;
				}
				else
				{
					V.wrong = (link->s[*link->p - 1] != ch);
					(*link->p)++;
				}
				break;

			case '\002':
				ch = link->rex[link->i - 1];
				link->i++;
				if (*link->p > strlen(link->s))
				{
					V.wrong = true;
				}
				else
				{
					V.wrong = (toupper(link->s[*link->p - 1]) != ch);
					(*link->p)++;
				}
				break;

			case '\003':
				n = link->rex[link->i - 1];
				link->i++;
				if (*link->p > strlen(link->s))
				{
					V.wrong = true;
				}
				else
				{
					if (link->cimode)
						ch1 = toupper(link->s[*link->p - 1]);
					else
						ch1 = link->s[*link->p - 1];
					while (n > 0 && ch1 > link->rex[link->i])
					{
						link->i += 2;
						n--;
					}
					if (n == 0 || ch1 < link->rex[link->i - 1])
						V.wrong = true;
					else
						(*link->p)++;
				}
				link->i += n + n;
				break;
				
			case '\004':
				if (*link->p > strlen(link->s))
				{
					V.wrong = true;
				}
				else
				{
					link->matchchar = link->s[*link->p - 1];
					(*link->p)++;
				}
				break;

			case '\005':
				if (*link->p != 1 && *link->p != strlen(link->s) + 1)
					V.wrong = true;
				break;

			case '\017':
				if ((*link->p != 1 && (link->s[*link->p - 2] == '_' || isalnum(link->s[*link->p - 2]))) ||
						(*link->p <= strlen(link->s) &&
						 !(link->s[*link->p - 1] == '_' || isalnum(link->s[*link->p - 1]))))
					V.wrong = true;
				break;

			case '\020':
				if (*link->p <= strlen(link->s) && (link->s[*link->p - 1] == '_' || isalnum(link->s[*link->p - 1])))
					V.wrong = true;
				break;

			case '\006':
				count__(0L, 255L, &V);
				break;

			case '\007':
				count__(0L, 1L, &V);
				break;

			case '\b':
				count__(1L, 255L, &V);
				break;

			case '\t':
				link->i += 2;
				count__((long)link->rex[link->i - 3], (long)link->rex[link->i - 2], &V);
				break;

			case '\n':
				V.wrong = comp(link);
				if (!V.wrong)
					skiprex(link->rex, link);
				break;

			case '\013':
				flag = false;
				link->i--;
				while (!strcmp((sprintf(STR2, "%c", link->rex[link->i - 1]), STR2), alternative_))
				{
					link->i++;
					if (!flag)
						flag = comp(link);
					skiprex(link->rex, link);
				}
				V.wrong = !flag;
				break;

			case '\f':
				if (!link->parflag)
				{
					for (j = 0; j <= 9; j++)
					{
						link->par[0][j] = -1;
						link->par[1][j] = -1;
					}
					link->parflag = true;
				}
				n = link->rex[link->i - 1];
				link->i++;
				link->par[0][n - 1] = *link->p;
				break;

			case '\015':
				n = link->rex[link->i - 1];
				link->i++;
				link->par[1][n - 1] = *link->p;
				break;

			case '\016':
				n = link->rex[link->i - 1];
				link->i++;
				if (link->parflag && link->par[1][n - 1] != -1)
				{
					j = link->par[1][n - 1] - link->par[0][n - 1];
					if (j + *link->p - 1 <= strlen(link->s))
					{
						sprintf(STR3, "%.*s", (int)j, link->s + *link->p - 1);
						sprintf(STR4, "%.*s", (int)j, link->s + link->par[0][n - 1] - 1);
						V.wrong = (strcmp(STR3, STR4) != 0);
						*link->p += j;
					}
					else
					{
						V.wrong = true;
					}
				} else
				{
					V.wrong = true;
				}
				break;

			case '\021':
				link->cimode = true;
				break;
		}
	}
	if (V.wrong)
	{
		link->i = savei;
		*link->p = savep;
	}
	return (!V.wrong);
}  /*comp*/

/* Local variables for replace: */
struct LOC_replace {
	struct LOC_compare *link;
	char temps[256];
	long notnum;
} ;

static void getrepl(struct LOC_replace *link)
{
	char ch;
	long j, n;
	char STR1[256], STR2[256], STR3[256];
	char FORLIM;

	while (strcmp((sprintf(STR1, "%c", link->link->rpl[link->link->i - 1]), STR1), end_))
	{
		ch = link->link->rpl[link->link->i - 1];
		link->link->i++;
		if (ch >= ' ')
		{
			sprintf(link->temps + strlen(link->temps), "%c", ch);
			continue;
		}
		switch (ch)
		{

			case '\001':
			case '\002':
				sprintf(link->temps + strlen(link->temps), "%c", link->link->rpl[link->link->i - 1]);
				link->link->i++;
				break;
				
			case '\003':
				n = link->link->rpl[link->link->i - 1];
				link->link->i++;
				while (n > 0)
				{
					FORLIM = link->link->rpl[link->link->i];
					for (ch = link->link->rpl[link->link->i - 1]; ch <= FORLIM; ch++)
						sprintf(link->temps + strlen(link->temps), "%c", ch);
					link->link->i += 2;
					n--;
				}
				break;
				
			case '\004':
				sprintf(link->temps + strlen(link->temps), "%c", link->link->matchchar);
				break;
				
			case '\005':
			case '\017':
			case '\020':
			case '\021':
				/* blank case */
				break;
				
			case '\006':
			case '\007':
			case '\b':
				skiprex(link->link->rpl, link->link);
				if (link->link->countright > 0)
					sprintf(link->temps + strlen(link->temps), "%.*s",(int)(link->link->countright - link->link->countleft),
							link->link->s + link->link->countleft - 1);
				break;
				
			case '\t':
				n = link->link->rpl[link->link->i - 1];
				j = link->link->i + 2;
				do
				{
					link->link->i = j;
					getrepl(link);
					n--;
				} while (n > 0);
				link->link->i++;
				break;
				
			case '\n':
				sprintf(STR3, "%c", link->link->rpl[link->link->i - 1]);
				if (!strcmp(STR3, nparen_) && !strcmp((sprintf(STR2, "%c", link->link->rpl[link->link->i + 1]), STR2), end_))
					link->notnum = link->link->rpl[link->link->i];
				skiprex(link->link->rpl, link->link);
				break;
				
			case '\f':
			case '\015':
				link->link->i++;
				break;
				
			case '\016':
				n = link->link->rpl[link->link->i - 1];
				link->link->i++;
				if (link->link->parflag && link->link->par[1][n - 1] != -1)
					sprintf(link->temps + strlen(link->temps), "%.*s",
							(int)(link->link->par[1][n - 1] - link->link->par[0][n - 1]),
							link->link->s + link->link->par[0][n - 1] - 1);
				break;
				
			case '\013':
				break;
				/* process up to end_, ignore other alternatives */
		}
	}
}  /*getrepl*/

static void replace(struct LOC_compare *link)
{
	struct LOC_replace V;
	char STR1[256];
	char STR2[256];

	V.link = link;
	link->i = 1;
	*V.temps = '\0';
	V.notnum = 0;
	getrepl(&V);
	if (V.notnum > 0 && link->parflag && link->par[1][V.notnum - 1] != -1)
	{
		sprintf(V.temps, "%s%s", strpart(STR1, link->s, (int)link->savep, (int)(link->par[0][V.notnum - 1] - 1)),
				strcpy(STR2, V.temps));
		strcat(V.temps, strpart(STR1, link->s, (int)link->par[1][V.notnum - 1], (int)(*link->p - 1)));
	}
	strcpy(link->s + link->savep - 1, link->s + *link->p - 1);
	sprintf(STR2, "%s%s", V.temps, link->s + link->savep - 1);
	strcpy(link->s + link->savep - 1, STR2);
	*link->p = link->savep + strlen(V.temps);
}  /*replace*/

static void compare(char *s_, char *rex_, char *rpl_, size_t *p_, int *found)
{
	struct LOC_compare V;

	V.s = s_;
	V.rex = rex_;
	V.rpl = rpl_;
	V.p = p_;
	V.i = 1;
	V.parflag = false;
	V.cimode = false;
	if (*V.rex == '\0')
	{
		*found = false;
		return;
	}
	if (*V.rpl == '\0')
	{
		*found = comp(&V);
		return;
	}
	V.countright = -1;
	V.matchchar = ' ';
	V.savep = *V.p;
	*found = comp(&V);
	if (*found)
		replace(&V);
}  /*compare*/

void re_search(char *s, char *rex, size_t *pos, size_t *last, size_t dir)
{
	int found, fast, vfast;
	char rpl[2];
	
	found = false;
	fast = (rex[0] >= ' ');
	vfast = (fast && dir == 1);
	rpl[0] = '\0';
	while (!found && *pos <= strlen(s) && *pos >= 1)
	{
		*last = *pos;
		if (!fast || s[*pos - 1] == rex[0])
		{
			compare(s, rex, rpl, last, &found);
			if (*last == *pos)   /*don't match null strings in search*/
				found = false;
		}
		if (found)
			break;
		if (vfast)
			*pos = strposc(s, rex[0], *pos + 1);
		else
			*pos += dir;
	}
	if (!found)
	{
		*pos = 0;
		*last = 0;
	} else
		(*last)--;
}

int re_compare(char *s, char *rex)
{
	size_t p;
	int found;
	char rpl[2];
	
	p = 1;
	rpl[0] = '\0';
	compare(s, rex, rpl, &p, &found);
	return (found && p == strlen(s) + 1);
}

void re_replace(char *s, char *rex, char *rpl, size_t *pos, size_t *last)
{
	int found;
	
	*last = *pos;
	compare(s, rex, rpl, last, &found);
	if (!found) {
		*pos = 0;
		*last = 0;
	} else
		(*last)--;
}

int re_matchstr(char *s_, char *pat)
{
	char s[256], rex[256];
	
	strcpy(s, s_);
	re_compile(pat, rex, "");
	return (re_compare(s, rex));
}

int re_matchsubstr(char *s_, char *pat, size_t start, size_t *last)
{
	char s[256], rex[256], rpl[256];
	
	strcpy(s, s_);
	re_compile(pat, rex, "");
	*rpl = '\0';
	re_replace(s, rex, rpl, &start, last);
	return (*last > 0);
}

void re_grepl(char *s, char *pat, char *rpl)
{
	char srex[256], rrex[256];
	size_t pos, last;

	pos = 1;
	re_compile(pat, srex, "");
	re_compile(rpl, rrex, "");
	do
	{
		re_search(s, srex, &pos, &last, 1L);
		if (pos > 0)
		{
			re_replace(s, srex, rrex, &pos, &last);
			pos = last + 1;
		}
	} while (pos != 0);
}


