/* "LOG", the circuit editing and simulation system,
   "DigLOG", the digital simulator for LOG.
   Copyright (C) 1985, 1990 David Gillespie.
   Author's address: daveg@csvax.caltech.edu; 256-80 Caltech/Pasadena CA 91125.

   "AnaLOG", the analog simulator for LOG.
   Copyright (C) 1985, 1990 John Lazzaro.
   Author's address: lazzaro@csvax.caltech.edu; 256-80 Caltech.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation (any version).

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   */

/*
   Logic Simulator   Version 4.1
   For HP 9836C

   Copyright (c)  1985

   David Gillespie
   */

#include <utils/p2c.h>
#include <utils/strings.h>
#include <graphics/newkbd.h>
#include <graphics/newcrt.h>
#include "logglobals.h"
#include "logfont.h"
#include "logstuff.h"
#include "utils.h"
#include "window.h"
#include "node.h"
#include "log.h"
#include "box.h"

short cursx;   ///< Current X position of cursor
short cursy;   ///< Current Y position of cursor
short cursx1;  ///< X position of wire starting point
short cursy1;  ///< Y position of wire starting point
short cursx2;  ///< X position of wire end point
short cursy2;  ///< Y position of wire end point
short chairx;  ///< X position of crosshair
short chairy;  ///< Y position of crosshair

int cursorflag;   ///< If cursor is displayed
int prevcursorflag; ///< If cursor was displayed before operation began

int chairflag;   ///< If crosshair is on window
int chairflag2;   ///< If crosshair will be on window
int rbandflag;   ///< If Rubber-band wire is on window

void initcolormap()
{
	m_setcolor((long)log_gray, 4L, 4L, 4L);
	m_setcolor((long)log_red, 15L, 0L, 0L);
	m_setcolor((long)log_green, 0L, 15L, 0L);
	m_setcolor((long)log_yellow, 15L, 15L, 0L);
	if (eightcolors) {
		m_setcolor((long)log_cred8, 15L, 0L, 0L);
		m_setcolor((long)log_black8, 0L, 0L, 0L);
	} else {
		m_setcolor((long)log_mred, 11L, 0L, 0L);
		m_setcolor((long)log_orange, 15L, 9L, 0L);
	}
	m_setcolor((long)log_cyan, 0L, 15L, 15L);
	m_setcolor((long)log_white, 15L, 15L, 15L);
	m_setcolor((long)log_black, 0L, 0L, 0L);
	m_setcolor((long)log_pink, 15L, 9L, 9L);
	m_setcolor((long)log_dyellow, 12L, 12L, 0L);
	m_setcolor((long)log_xgreen, 0L, 13L, 0L);
	m_setcolor((long)log_dcyan, 0L, 10L, 10L);
	m_setcolor((long)log_dred, 8L, 0L, 0L);
	m_setcolor((long)log_lgray, 10L, 10L, 10L);
	m_setcolor((long)log_cred, 15L, 0L, 0L);
	m_vseecolors(0L, 16L, rcolormap, gcolormap, bcolormap);
}


void fixcolormap()
{
	m_vsetcolors(0L, 16L, rcolormap, gcolormap, bcolormap);
	recolor_log_cursors(gg.color.cursor, true);
}


/// Initialize color graphics, graphics tablet, low-level flags & variables.
void initwindow()
{
	cursorflag = false;
	m_init_screen();
	m_upside_down();
	m_setlinestyle(1L, 0xf0f0L);
	m_setlinestyle(2L, 0xaaaaL);
	m_setfont(logfont_lfont);
	init_X_window();
	onescreen = nc_alphashared();
	eightcolors = (m_maxcolor <= 7);

	txacross = nc_curWindow->width - 1;
	txdown = nc_curWindow->height - 1;
	across = m_across;
	down = m_down;
	baseline = down - 53;   /* Position of baseline on window */
	line1 = down - 43;   /* Position of first text line in menu */
	line2 = down - 23;   /* Position of second text line in menu */
	histdown = down - 26;
	histdivsacross = (double)(across - histleft) / histdivision;
	histvalrange = (double)histdown / histdivision * 5;
	kindgroupsize = (across - 160) / kindgroupspacing;
	kindgroupleft = (across - kindgroupsize * kindgroupspacing) / 2;
	kindgroupright = kindgroupleft + kindgroupsize * kindgroupspacing;
	kindgroupstart = kindgroupleft + kindgroupspacing / 2;
	kindgroupbase = (baseline + down) / 2;

	nc_printf("\200\210");

	cursorflag = false;
	gg.stillnear = true;
}


void initwindow2()
{
	initwindow();
	m_clear();
	m_graphics_on();
}

/// Turn off the stupid *blinking* Alpha cursor.
void noblink()
{
	nc_cursXY(-1, -1);
	/*  oldx := xpos;
oldy := ypos;
xpos := -1;
ypos := -1;
call(updatecursorhook);
xpos := oldx;
ypos := oldy; */
}


void clearalpha()
{
	if (!onescreen)
	{
		nc_printf("\f");
		noblink();
		m_alpha_off();
	}
	nc_gotoXY(0, 0);
}


void showalpha()
{
	m_graphics_off();
	m_alpha_on();
}

/// Draw a line in grid coordinates.
void line(short x1, short y1, short x2, short y2)
{
	m_drawline(x1 * gg.scale - gg.xoff, y1 * gg.scale - gg.yoff,
			x2 * gg.scale - gg.xoff, y2 * gg.scale - gg.yoff);
}


/// Draw a horizontal line in grid coordinates.
void hline(short x1, short x2, short y)
{
	short yy;

	yy = y * gg.scale - gg.yoff;
	m_drawline(x1 * gg.scale - gg.xoff, (long)yy, x2 * gg.scale - gg.xoff, (long)yy);
}


/// Draw a vertical line in grid coordinates.
void vline(short x, short y1, short y2)
{
	short xx;

	xx = x * gg.scale - gg.xoff;
	m_drawline((long)xx, y1 * gg.scale - gg.yoff, (long)xx,
			y2 * gg.scale - gg.yoff);
}


/// Draw a point in grid coordinates.
void point(short x, short y)
{
	m_drawpoint(x * gg.scale - gg.xoff, y * gg.scale - gg.yoff);
}


/// Draw a rectangle in grid coordinates.
void rect(short x1, short y1, short x2, short y2)
{
	m_drawrect(x1 * gg.scale - gg.xoff, y1 * gg.scale - gg.yoff,
			x2 * gg.scale - gg.xoff, y2 * gg.scale - gg.yoff);
}

/// Draw a filled rectangle in grid coords.
void box(short x1, short y1, short x2, short y2)
{
	m_fillrect(x1 * gg.scale - gg.xoff, y1 * gg.scale - gg.yoff,
			x2 * gg.scale - gg.xoff, y2 * gg.scale - gg.yoff);
}

/// Draw a text string.
void drawstr2(int x, int y, char *s)
{
	long save;

	if (m_curcolor() < 16)
	{
		save = m_curcolor();
		m_color((long)gg.color.backgr);
		m_fillrect(x - 1L, y - 2L, x + m_strwidth(logfont_lfont, s), y + 9L);
		m_color(save);
	}
	m_move((long)x, (long)y);
	m_displaytext(s);
}


void rightstr2(int x, int y, char *s)
{
	long save;

	if (m_curcolor() < 16)
	{
		save = m_curcolor();
		m_color((long)gg.color.backgr);
		m_fillrect(x + m_strwidth(logfont_lfont, s) - 1, y - 2L, x + 1L, y + 9L);
		m_color(save);
	}
	m_rightstr((long)x, (long)y, logfont_lfont, s);
}


void centerstr2(int x, int y, char *s)
{
	short w;
	long save;

	if (m_curcolor() < 16)
	{
		save = m_curcolor();
		m_color((long)gg.color.backgr);
		w = m_strwidth(logfont_lfont, s) / 2 + 2;
		m_fillrect((long)(x - w), y - 2L, (long)(x + w), y + 9L);
		m_color(save);
	}
	m_centerstr((long)x, (long)y, logfont_lfont, s);
}


/// Draw a single digit at the specified position (window coordinates).
void drawnum1(short x, short y, short n)
{
	char s[2];

	m_move((long)x, (long)y);
	strcpy(s, " ");
	s[0] = n + '0';
	m_displaytext(s);
}


/// Draw a two-digit number at the specified position (window coordinates).
void drawnum2(short x, short y, short n)
{
	char s[3];

	m_move((long)x, (long)y);
	strcpy(s, "  ");
	if (n >= 10)
		s[0] = n / 10 + '0';
	s[1] = n % 10 + '0';
	m_displaytext(s);
}


void xformcoords(log_grec *g, short *x, short *y)
{
	short xx, yy;

	if (g == NULL)
	{
		*x = *x * gg.scale - gg.xoff;
		*y = *y * gg.scale - gg.yoff;
		return;
	}
	xx = *x * gg.scale;
	yy = *y * gg.scale;
	*x = g->x * gg.scale - gg.xoff +
		(log_rotxx[g->rot] * xx + log_rotyx[g->rot] * yy) / log_scale0;
	*y = g->y * gg.scale - gg.yoff +
		(log_rotxy[g->rot] * xx + log_rotyy[g->rot] * yy) / log_scale0;
}


void plainxformcoords(log_grec *g, short *x, short *y)
{
	short xx;

	if (g == NULL)
	{
		*x *= log_scale0;
		*y *= log_scale0;
		return;
	}
	xx = *x;
	*x = g->x * log_scale0 + log_rotxx[g->rot] * xx + log_rotyx[g->rot] * *y;
	*y = g->y * log_scale0 + log_rotxy[g->rot] * xx + log_rotyy[g->rot] * *y;
}


/// Allow drawing only in the working area.
void clipon()
{
	m_clip(0L, 0L, (long)across, baseline - 1L);
}


/// Allow drawing in working area and menus.
void clipoff()
{
	m_noclip();
}

/// Erase a rectangular area of the window using window coordinates.
void uerase(short x1, short y1, short x2, short y2)
{
	m_color((long)gg.color.backgr);
	m_fillrect((long)x1, (long)y1, (long)x2, (long)y2);
}


/// Erase a rectangular area of the window using grid coordinates.
void erase(short x1, short y1, short x2, short y2)
{
	m_color((long)gg.color.backgr);
	box(x1, y1, x2, y2);
}

/** @brief Attempt to force a line to be horizontal
			or vertical.  If possible, second pair
			of coordinates is changed appropriately.

	@returns true if possible, false if not.
*/
int hvline(short x1, short y1, short *x2, short *y2)
{
	int Result;

	Result = true;
	if (x1 == *x2 && y1 == *y2)
		return false;
	if (abs(*x2 - x1) > abs(*y2 - y1) && abs(*y2 - y1) < abs(*x2 - x1) / 3 + 20)
	{   /*7/14/88*/
		*y2 = y1;
		return Result;
	}
	if (abs(*y2 - y1) <= abs(*x2 - x1) ||
			abs(*x2 - x1) >= abs(*y2 - y1) / 3 + 20)
		/*7/14/88*/
		return false;
	*x2 = x1;
	return Result;
}

/// Adjust window coordinates so that they are aligned to the nearest grid point.
void fixxy(short *x, short *y)
{
	*x = (*x + gg.xoff + gg.hscale) / gg.scale * gg.scale - gg.xoff;
	*y = (*y + gg.yoff + gg.hscale) / gg.scale * gg.scale - gg.yoff;
}

/// Turn off crosshair (no effect until next call of CURSOR).
void nocrosshair()
{
	chairflag2 = false;
}


/// Move the crosshair to a particular position in grid coordinates.
void crosshair(short x, short y)
{
	chairx = x * gg.scale - gg.xoff;
	chairy = y * gg.scale - gg.yoff;
	chairflag2 = true;
}

/** @brief Draw the cursor on the window.
			The exact shape drawn depends upon the
			current mode.
			If the cursor is already on the window, it
			is removed.

	Possible cursors:
		- Arrow (normal).
		- Probe (if Logic Probe mode).
		- Big crosshair (if "Grid" mode on).
		- Long arrow (if Copy command).
		- Scissors (if Delete command).
		- Rectangle (if Paste command).
		- Green line (if drawing a wire).
*/
void xorcursor()
{
	long curcm;
	short cx0, cy0;

	gg.cx_min = 0;
	gg.cx_max = across;
	gg.cy_min = 0;
	gg.cy_max = baseline - 1;
	curcm = m_curcolormode();
	m_colormode((long)m_xor);
	m_color((long)gg.color.cursor);
	if (!chairflag || cursortype == grid)
	{
		if (cursortype != normal && cursy < baseline)
		{
			switch (cursortype)
			{

				case grid:
					choose_log_cursor(0);
					if (chairflag) {
						m_drawline(cursx1 - 3L, cursy1 - 5L, cursx1 + 3L, cursy1 + 5L);
						m_drawline(cursx1 - 3L, cursy1 + 5L, cursx1 + 3L, cursy1 - 5L);
					}
					cx0 = (cursx + gg.xoff) / gg.scale * gg.scale - gg.xoff;
					cy0 = (cursy + gg.yoff) / gg.scale * gg.scale - gg.yoff;
					m_drawline(0L, (long)cy0, (long)across, (long)cy0);
					m_drawline((long)cx0, 0L, (long)cx0, baseline - 1L);
					break;

				case delete__:
					choose_log_cursor(2);
					break;

				case copy_:
					choose_log_cursor(1);
					break;

				case boxcursor:
					choose_log_cursor(4);
					break;

				case paste:
					choose_log_cursor(0);
					cx0 = (cursx + gg.xoff + 2) / gg.scale;
					cy0 = (cursy + gg.yoff + 2) / gg.scale;
					m_saveclip();
					m_clip(0L, 0L, (long)across, baseline - 1L);
					gg.cx_min = cx0;
					gg.cx_max = cx0 + copybuf.dxcopy;
					gg.cy_min = cy0;
					gg.cy_max = cy0 + copybuf.dycopy;
					rect(gg.cx_min, gg.cy_min, gg.cx_max, gg.cy_max);
					m_unclip();
					break;

				default:
					break;
			}
		}
		else
		{
			if (gg.probemode)
			{
				choose_log_cursor(3);
			}
			else
			{
				choose_log_cursor(0);
			}
		}
	}
	if (chairflag)
	{
		if (rbandflag)
		{
			choose_log_cursor(0);
			if (vlsi)
				m_color((long)gg.color.wire[curwcolor - log_wcol_normal]);
			else
				m_color((long)gg.color.xwire);
			m_drawline((long)cursx1, (long)cursy1, (long)cursx2, (long)cursy2);
		}
		else
		{
			choose_log_cursor(0);
		}
	}
	oldcursortype = cursortype;
	m_colormode(curcm);

	/*and (cursy < baseline)*/
}


/// Used to remove cursor from window while drawing graphics.
void hidecursor()
{
	if (cursorflag)
		xorcursor();
}


void hidecursor_hook()
{
	if (!cursorhide)
		hidecursor();
	cursorhide = true;
}


void unhidecursor_hook()
{
	if (cursorhide)
		hidecursor();
	cursorhide = false;
}


void hidecursorrect_hook(long x1, long y1, long x2, long y2)
{
	if (!cursorhide && x1 <= gg.cx_max && x2 >= gg.cx_min && y1 <= gg.cy_max &&
			y2 >= gg.cy_min)
		hidecursor_hook();
}

/// Position the cursor to a particular point (window coordinates).
void drawcursor(short x, short y)
{
	if (x != cursx || y != cursy || !cursorflag || cursortype != oldcursortype)
	{
		if (cursorflag && !cursorhide)
			xorcursor();
		cursx = x;
		cursy = y;
		cursx1 = chairx;
		cursy1 = chairy;
		cursx2 = x;
		cursy2 = y;
		chairflag = chairflag2;
		if (chairflag)
		{
			chairflag = true;
			fixxy(&cursx2, &cursy2);
			rbandflag = (abs(cursx1 - cursx2) + abs(cursy1 - cursy2) >= gg.scale &&
					hvline(cursx1, cursy1, &cursx2, &cursy2));
		}
		xorcursor();
	}
	else
	{
		if (!cursorflag && !cursorhide)
			xorcursor();
	}
	cursorflag = true;
	cursorhide = false;
}

/// Take the cursor off the window.
void remcursor()
{
	prevcursorflag = cursorflag;
	if (cursorflag)
		xorcursor();
	cursorflag = false;
	cursorhide = false;
}

/// Restore the cursor to its previous on/off status.
void restorecursor()
{
	if (prevcursorflag != cursorflag)
		xorcursor();
	cursorflag = prevcursorflag;
}

/// Clear the window.
void clearwindow()
{
	remcursor();
	m_noclip();
	m_vsetcolors((long)gg.color.backgr, 1L, &rcolormap[gg.color.backgr],
			&gcolormap[gg.color.backgr], &bcolormap[gg.color.backgr]);
	m_color((long)gg.color.backgr);
	m_fillrect(0L, 0L, (long)across, (long)down);
	fixcolormap();
	resetmessages();
	bottomcount = 0;
	gg.showpage = 0;
}


void clearshowalpha()
{
	if (onescreen)
		clearwindow();
	nc_printf("\f");
	nc_gotoXY(0, 0);
	noblink();
	showalpha();
}


void drawvector2(log_vectorrec *v, long x1, long y1, long x2, long y2)
{
	long diam, savecol, coord;

	switch (v->vkind)
	{

		case 'v':
			m_drawline(x1, y1, x2, y2);
			break;

		case 'C':
			diam = P_imax2(labs(x1 - x2), labs(y1 - y2));
			switch (diam)
			{

				case 3:
					if (x1 == x2)
					{
						coord = P_imin2(y1, y2);
						m_move2(x1, coord);
						m_draw(x1 + 2, coord + 1);
						m_draw(x1 + 2, coord + 2);
						m_draw(x1, coord + 3);
						m_draw(x1 - 2, coord + 2);
						m_draw(x1 - 2, coord + 1);
						m_draw(x1, coord);
					}
					else if (y1 == y2)
					{
						coord = P_imin2(x1, x2);
						m_move2(coord, y1);
						m_draw(coord + 1, y1 + 2);
						m_draw(coord + 2, y1 + 2);
						m_draw(coord + 3, y1);
						m_draw(coord + 2, y1 - 2);
						m_draw(coord + 1, y1 - 2);
						m_draw(coord, y1);
					}
					else
					{
						m_circle((x1 + x2) / 2, (y1 + y2) / 2, 1L);
					}
					break;

				case 5:
					if (x1 == x2)
					{
						coord = P_imin2(y1, y2);
						m_move2(x1, coord);
						m_draw(x1 + 3, coord + 1);
						m_draw(x1 + 3, coord + 4);
						m_draw(x1, coord + 5);
						m_draw(x1 - 3, coord + 4);
						m_draw(x1 - 3, coord + 1);
						m_draw(x1, coord);
					}
					else if (y1 == y2)
					{
						coord = P_imin2(x1, x2);
						m_move2(coord, y1);
						m_draw(coord + 1, y1 + 3);
						m_draw(coord + 4, y1 + 3);
						m_draw(coord + 5, y1);
						m_draw(coord + 4, y1 - 3);
						m_draw(coord + 1, y1 - 3);
						m_draw(coord, y1);
					}
					else
					{
						m_circle((x1 + x2) / 2, (y1 + y2) / 2, 2L);
					}
					break;

				default:
					m_circle((x1 + x2) / 2, (y1 + y2) / 2, diam / 2);
					break;
			}
			break;

		case 'e':
			savecol = m_curcolor();
			m_color(m_trans);
			m_ellipse((x1 + x2) / 2, (y1 + y2) / 2, (x1 - x2) / 2, (y1 - y2) / 2,
					savecol);
			m_color(savecol);
			break;

		case 'E':
			m_ellipse((x1 + x2) / 2, (y1 + y2) / 2, (x1 - x2) / 2, (y1 - y2) / 2,
					m_curcolor());
			break;

		case 'r':   /*should do a true roundrect*/
			m_drawrect(x1, y1, x2, y2);
			break;
	}
}


void drawvector4(log_vectorrec *v, long x1, long y1, long x2, long y2, long x3, long y3, long x4, long y4)
{
	long xarr[4], yarr[4];

	switch (v->vkind)
	{

		case 'c':
			m_bezier(x1, y1, x2, y2, x3, y3, x4, y4);
			break;

		case 'F':
			xarr[0] = x1;
			yarr[0] = y1;
			xarr[1] = x2;
			yarr[1] = y2;
			xarr[2] = x3;
			yarr[2] = y3;
			xarr[3] = x4;
			yarr[3] = y4;
			m_fillpoly(4L, (int *) xarr, (int *) yarr);
			break;
	}
}


#define big1            "0.,+123-456*789/\0()^ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz "
#define big2            "0.,\000123-456\000789 "


void drawtext(log_vectorrec *v, long scale, long rot, long xx, long yy)
{
	long sc, i, j, tx, ty;
	char mode;
	char buf[256];
	long FORLIM;

	sc = v->UU.U116.tsize * scale;
	if (sc >= log_scale0 * 70)
	{
		tx = (v->UU.U116.torg % 3 - 1) * log_rotxx[rot] +
			(v->UU.U116.torg / 3 - 1) * log_rotyx[rot];
		ty = (v->UU.U116.torg % 3 - 1) * log_rotxy[rot] +
			(v->UU.U116.torg / 3 - 1) * log_rotyy[rot];
		i = strposc(v->UU.U116.sp, '$', 1L);
		if (i > 0 && i < strlen(v->UU.U116.sp))
		{
			strcpy(buf, v->UU.U116.sp);
			mode = 'X';
			while (i <= strlen(buf)) {
				if (buf[i - 1] == '$' && i < strlen(buf))
				{
					switch (toupper(buf[i]))
					{

						case '$':
							strcpy_overlap(buf + i - 1, buf + i);
							i++;
							break;

						case 'P':
						case 'B':
						case 'O':
						case 'X':
							mode = toupper(buf[i]);   /*close enough*/
							strcpy_overlap(buf + i - 1, buf + i + 1);
							break;

						default:
							i++;
							break;
					}
					continue;
				}
				switch (mode)
				{

					case 'P':
						j = strposc("0.,+123-456*789/E()^", buf[i - 1], 1L);
						if (j > 0)
							buf[i - 1] = (char)(j + 147);
						break;

					case 'B':
						j = strposc("0.,+123-456*789/\0()^ABCDEFGHIJKLMNOPQRSTUVWXYZ",
								toupper(buf[i - 1]), 1L);
						if (j > 20)
							buf[i - 1] = (char)(j + 147);
						else if (j > 0)
							buf[i - 1] = (char)(j + 127);
						break;

					case 'O':
						j = strposc("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789",
								toupper(buf[i - 1]), 1L);
						if (j > 0)
							buf[i - 1] = (char)(j + 213);
						break;
				}
				i++;
			}
			switch (tx)
			{

				case -1:
					m_drawstr(xx, yy - (ty + 1) * 3, NULL, buf);
					break;

				case 0:
					m_centerstr(xx, yy - (ty + 1) * 3, NULL, buf);
					break;

				case 1:
					m_rightstr(xx, yy - (ty + 1) * 3, NULL, buf);
					break;
			}
			return;
		}
		switch (tx)
		{

			case -1:
				m_drawstr(xx, yy - (ty + 1) * 3, NULL, v->UU.U116.sp);
				break;

			case 0:
				m_centerstr(xx, yy - (ty + 1) * 3, NULL, v->UU.U116.sp);
				break;

			case 1:
				m_rightstr(xx, yy - (ty + 1) * 3, NULL, v->UU.U116.sp);
				break;
		}
		return;
	}
	if (sc < log_scale0 * 40)
		return;
	if (sc >= log_scale0 * 50) {
		if (!strsubset(v->UU.U116.sp, big1))
			return;
		buf[strlen(v->UU.U116.sp)] = '\0';
		FORLIM = strlen(v->UU.U116.sp);
		for (i = 0; i < FORLIM; i++) {
			j = strposc(big1, v->UU.U116.sp[i], 1L);
			if (j <= 20)
				buf[i] = (char)(j + 127);
			else if (j <= 46)
				buf[i] = (char)(j + 147);
			else if (j <= 72)
				buf[i] = (char)(j + 121);
			else
				buf[i] = ' ';
		}
		tx = (v->UU.U116.torg % 3 - 1) * log_rotxx[rot] +
			(v->UU.U116.torg / 3 - 1) * log_rotyx[rot];
		ty = (v->UU.U116.torg % 3 - 1) * log_rotxy[rot] +
			(v->UU.U116.torg / 3 - 1) * log_rotyy[rot];
		switch (tx)
		{

			case -1:
				m_drawstr(xx, yy - (ty + 2) * 2, NULL, buf);
				break;

			case 0:
				m_centerstr(xx, yy - (ty + 2) * 2, NULL, buf);
				break;

			case 1:
				m_rightstr(xx, yy - (ty + 2) * 2, NULL, buf);
				break;
		}
		return;
	}
	if (!strsubset(v->UU.U116.sp, big2))
		return;
	buf[strlen(v->UU.U116.sp)] = '\0';
	FORLIM = strlen(v->UU.U116.sp);
	for (i = 0; i < FORLIM; i++)
	{
		j = strposc(big2, v->UU.U116.sp[i], 1L);
		if (j <= 15)
			buf[i] = (char)(j + 193);
		else
			buf[i] = ' ';
	}
	tx = (v->UU.U116.torg % 3 - 1) * log_rotxx[rot] +
		(v->UU.U116.torg / 3 - 1) * log_rotyx[rot];
	ty = (v->UU.U116.torg % 3 - 1) * log_rotxy[rot] +
		(v->UU.U116.torg / 3 - 1) * log_rotyy[rot];
	switch (tx)
	{

		case -1:
			m_drawstr(xx, yy - (ty + 1) * 2 - 3, NULL, buf);
			break;

		case 0:
			m_centerstr(xx, yy - (ty + 1) * 2 - 3, NULL, buf);
			break;

		case 1:
			m_rightstr(xx, yy - (ty + 1) * 2 - 3, NULL, buf);
			break;
	}
}

#undef big1
#undef big2


void plaindrawgatec(short x, short y, short g, short c)
{
	short i, z, rxx, rxy, ryx, ryy;
	log_krec *k;
	int hascolor;
	short FORLIM;
	log_vectorrec *WITH1;

	k = kind[(g & (log_kindoffset - 1)) - 1];
	z = log_scale0 * k->bbmax + 2;
	if (x < -z || x > across + z || y < -z || y > down + z)
		return;
	hidecursor();
	hascolor = (c != gg.color.backgr && c != gg.color.conflict);
	m_color((long)c);
	if (g < log_kindoffset)
	{
		FORLIM = k->numvects;
		for (i = 0; i < FORLIM; i++)
		{
			WITH1 = &k->vector[i];
			switch (WITH1->vkind)
			{

				case '1':
					if (hascolor)
						m_color((long)c);
					break;

				case '2':
					if (hascolor)
						m_color((long)gategreen);
					break;

				case '3':
					if (hascolor)
						m_color((long)gateyellow);
					break;

				case '4':
					if (hascolor)
						m_color((long)gatered);
					break;

				case '5':
					if (hascolor)
						m_color((long)gateorange);
					break;

				case '6':
					if (hascolor)
						m_color((long)gateblack);
					break;

				case '7':
					if (hascolor)
						m_color((long)gatewhite);
					break;

				case 'v':
				case 'C':
				case 'e':
				case 'E':
				case 'r':
					drawvector2(&k->vector[i], (long)(x + WITH1->x1),
							(long)(y + WITH1->y1), (long)(x + WITH1->UU.U99.x2),
							(long)(y + WITH1->UU.U99.y2));
					break;

				case 'c':
				case 'F':
					drawvector4(&k->vector[i], (long)(x + WITH1->x1),
							(long)(y + WITH1->y1), (long)(x + WITH1->UU.U99.x2),
							(long)(y + WITH1->UU.U99.y2), (long)(x + WITH1->UU.U99.x3),
							(long)(y + WITH1->UU.U99.y3), (long)(x + WITH1->UU.U99.x4),
							(long)(y + WITH1->UU.U99.y4));
					break;

				case 't':
					drawtext(&k->vector[i], (long)log_scale0, 0L, (long)(x + WITH1->x1),
							(long)(y + WITH1->y1));
					break;
			}
		}
	}
	else
	{
		z = g / log_kindoffset;
		rxx = log_rotxx[z];
		rxy = log_rotxy[z];
		ryx = log_rotyx[z];
		ryy = log_rotyy[z];
		FORLIM = k->numvects;
		for (i = 0; i < FORLIM; i++)
		{
			WITH1 = &k->vector[i];
			switch (WITH1->vkind)
			{

				case '1':
					if (hascolor)
						m_color((long)c);
					break;

				case '2':
					if (hascolor)
						m_color((long)gategreen);
					break;

				case '3':
					if (hascolor)
						m_color((long)gateyellow);
					break;

				case '4':
					if (hascolor)
						m_color((long)gatered);
					break;

				case '5':
					if (hascolor)
						m_color((long)gateorange);
					break;

				case '6':
					if (hascolor)
						m_color((long)gateblack);
					break;

				case '7':
					if (hascolor)
						m_color((long)gatewhite);
					break;

				case 'v':
				case 'C':
				case 'e':
				case 'E':
				case 'r':
					drawvector2(&k->vector[i],
							(long)(x + WITH1->x1 * rxx + WITH1->y1 * ryx),
							(long)(y + WITH1->x1 * rxy + WITH1->y1 * ryy),
							(long)(x + WITH1->UU.U99.x2 * rxx + WITH1->UU.U99.y2 * ryx),
							(long)(y + WITH1->UU.U99.x2 * rxy + WITH1->UU.U99.y2 * ryy));
					break;

				case 'c':
				case 'F':
					drawvector4(&k->vector[i],
							(long)(x + WITH1->x1 * rxx + WITH1->y1 * ryx),
							(long)(y + WITH1->x1 * rxy + WITH1->y1 * ryy),
							(long)(x + WITH1->UU.U99.x2 * rxx + WITH1->UU.U99.y2 * ryx),
							(long)(y + WITH1->UU.U99.x2 * rxy + WITH1->UU.U99.y2 * ryy),
							(long)(x + WITH1->UU.U99.x3 * rxx + WITH1->UU.U99.y3 * ryx),
							(long)(y + WITH1->UU.U99.x3 * rxy + WITH1->UU.U99.y3 * ryy),
							(long)(x + WITH1->UU.U99.x4 * rxx + WITH1->UU.U99.y4 * ryx),
							(long)(y + WITH1->UU.U99.x4 * rxy + WITH1->UU.U99.y4 * ryy));
					break;

				case 't':
					drawtext(&k->vector[i], (long)log_scale0, (long)z,
							(long)(x + WITH1->x1 * rxx + WITH1->y1 * ryx),
							(long)(y + WITH1->x1 * rxy + WITH1->y1 * ryy));
					break;
			}
		}
	}
	hidecursor();
}

/// Draw a gate of specified color at a certain position (window coordinates).
void udrawgatec(int x, int y, int g, int c)
{
	short i, z, rxx, rxy, ryx, ryy;
	log_krec *k;
	int hascolor;
	short FORLIM;
	log_vectorrec *WITH1;

	if (g == 0)
	{
		hidecursor();
		m_color((long)c);
		m_drawpoint((long)x, (long)y);
		hidecursor();
		return;
	}
	k = kind[(g & (log_kindoffset - 1)) - 1];
	if (k->bbmax <= maxsafescale)
	{
		plaindrawgatec(x, y, g, c);
		return;
	}
	hascolor = (c != gg.color.backgr);
	z = log_scale0 * k->bbmax + 2;
	if (x < -z || x > across + z || y < -z || y > down + z)
		return;
	z = g / log_kindoffset;
	rxx = log_rotxx[z] * maxsafescale;
	rxy = log_rotxy[z] * maxsafescale;
	ryx = log_rotyx[z] * maxsafescale;
	ryy = log_rotyy[z] * maxsafescale;
	hidecursor();
	m_color((long)c);
	FORLIM = k->numvects;
	for (i = 0; i < FORLIM; i++)
	{
		WITH1 = &k->vector[i];
		switch (WITH1->vkind)
		{

			case '1':
				if (hascolor)
					m_color((long)c);
				break;

			case '2':
				if (hascolor)
					m_color((long)gategreen);
				break;

			case '3':
				if (hascolor)
					m_color((long)gateyellow);
				break;

			case '4':
				if (hascolor)
					m_color((long)gatered);
				break;

			case '5':
				if (hascolor)
					m_color((long)gateorange);
				break;

			case '6':
				if (hascolor)
					m_color((long)gateblack);
				break;

			case '7':
				if (hascolor)
					m_color((long)gatewhite);
				break;

			case 'v':
			case 'C':
			case 'e':
			case 'E':
			case 'r':
				drawvector2(&k->vector[i],
						(long)(x + (WITH1->x1 * rxx + WITH1->y1 * ryx) / k->bbmax),
						(long)(y + (WITH1->x1 * rxy + WITH1->y1 * ryy) / k->bbmax),
						(long)(x + (WITH1->UU.U99.x2 * rxx + WITH1->UU.U99.y2 * ryx) / k->bbmax),
						(long)(y + (WITH1->UU.U99.x2 * rxy + WITH1->UU.U99.y2 * ryy) / k->bbmax));
				break;

			case 'c':
			case 'F':
				drawvector4(&k->vector[i], (long)(x + (WITH1->x1 * rxx +
								WITH1->y1 * ryx) / k->bbmax), (long)
						(y + (WITH1->x1 * rxy + WITH1->y1 * ryy) / k->bbmax), (long)
						(x + (WITH1->UU.U99.x2 * rxx + WITH1->UU.U99.y2 * ryx) / k->bbmax),
						(long)
						(y + (WITH1->UU.U99.x2 * rxy + WITH1->UU.U99.y2 * ryy) / k->bbmax),
						(long)
						(x + (WITH1->UU.U99.x3 * rxx + WITH1->UU.U99.y3 * ryx) / k->bbmax),
						(long)
						(y + (WITH1->UU.U99.x3 * rxy + WITH1->UU.U99.y3 * ryy) / k->bbmax),
						(long)
						(x + (WITH1->UU.U99.x4 * rxx + WITH1->UU.U99.y4 * ryx) / k->bbmax),
						(long)
						(y + (WITH1->UU.U99.x4 * rxy + WITH1->UU.U99.y4 * ryy) / k->bbmax));
				break;
		}
	}
	hidecursor();
}


void drawgatedotsc(short x, short y, short g, short c)
{
	short i, z, rxx, rxy, ryx, ryy;
	log_krec *WITH;
	short FORLIM;

	if (g == 0 || vlsi)
		return;
	m_color((long)c);
	z = g / log_kindoffset;
	if (z != 0)
	{
		rxx = log_rotxx[z];
		rxy = log_rotxy[z];
		ryx = log_rotyx[z];
		ryy = log_rotyy[z];
		WITH = kind[(g & (log_kindoffset - 1)) - 1];
		FORLIM = WITH->numpins;
		for (i = 0; i < FORLIM; i++)
			point(x + WITH->pin[i].x * rxx + WITH->pin[i].y * ryx,
					y + WITH->pin[i].x * rxy + WITH->pin[i].y * ryy);
		return;
	}
	WITH = kind[(g & (log_kindoffset - 1)) - 1];
	FORLIM = WITH->numpins;
	for (i = 0; i < FORLIM; i++)
		point(x + WITH->pin[i].x, y + WITH->pin[i].y);
}

/// Draw a gate of specified color at a certain position (grid coordinates).
void drawgatec(int x, int y, int g, int c)
{
	short i, xx, yy, z, rxx, rxy, ryx, ryy;
	long savecol;
	int hascolor;
	log_krec *WITH;
	short FORLIM;
	log_vectorrec *WITH1;

	if (zoom == 0)
	{
		plaindrawgatec((int)(x * gg.scale - gg.xoff),
				(int)(y * gg.scale - gg.yoff), g, c);
	}
	else
	{
		z = g / log_kindoffset;
		rxx = log_rotxx[z] * gg.scale;
		rxy = log_rotxy[z] * gg.scale;
		ryx = log_rotyx[z] * gg.scale;
		ryy = log_rotyy[z] * gg.scale;
		xx = x * gg.scale - gg.xoff;
		yy = y * gg.scale - gg.yoff;
		hidecursor();
		m_color((long)c);
		hascolor = (c != gg.color.backgr && c != gg.color.conflict);
		WITH = kind[(g & (log_kindoffset - 1)) - 1];
		FORLIM = WITH->numvects;
		for (i = 0; i < FORLIM; i++)
		{
			WITH1 = &WITH->vector[i];
			switch (WITH1->vkind)
			{

				case '1':
					if (hascolor)
						m_color((long)c);
					break;

				case '2':
					if (hascolor)
						m_color((long)gategreen);
					break;

				case '3':
					if (hascolor)
						m_color((long)gateyellow);
					break;

				case '4':
					if (hascolor)
						m_color((long)gatered);
					break;

				case '5':
					if (hascolor)
						m_color((long)gateorange);
					break;

				case '6':
					if (hascolor)
						m_color((long)gateblack);
					break;

				case '7':
					if (hascolor)
						m_color((long)gatewhite);
					break;

				case 'v':
				case 'r':
					drawvector2(
							&WITH->vector[i],
							(long)(xx + (WITH1->x1 * rxx + WITH1->y1 * ryx) / log_scale0),
							(long)(yy + (WITH1->x1 * rxy + WITH1->y1 * ryy) / log_scale0),
							(long)
							(xx + (WITH1->UU.U99.x2 * rxx + WITH1->UU.U99.y2 * ryx) / log_scale0),
							(long)
							(yy +
							 (WITH1->UU.U99.x2 * rxy + WITH1->UU.U99.y2 * ryy) / log_scale0));
					break;

				case 'C':
					m_circle(xx + ((WITH1->x1 + WITH1->UU.U99.x2) * rxx + (WITH1->y1 +
									WITH1->UU.U99.y2) * ryx) / (log_scale0 * 2L),
							yy + ((WITH1->x1 + WITH1->UU.U99.x2) * rxy + (WITH1->y1 + WITH1->
									UU.U99.y2) * ryy) / (log_scale0 * 2L),
							P_imax2((long)abs((WITH1->x1 - WITH1->UU.U99.x2) * rxx +
									(WITH1->y1 - WITH1->UU.U99.y2) * ryx),
								(long)abs((WITH1->x1 - WITH1->UU.U99.x2) * ryx +
									(WITH1->y1 - WITH1->UU.U99.y2) * ryy)) /
							(log_scale0 * 2));
					break;

				case 'e':
					savecol = m_curcolor();
					m_color(m_trans);
					m_ellipse(
							xx + ((WITH1->x1 + WITH1->UU.U99.x2) * rxx +
								(WITH1->y1 + WITH1->UU.U99.y2) * ryx) / (log_scale0 * 2L),
							yy + ((WITH1->x1 + WITH1->UU.U99.x2) * rxy +
								(WITH1->y1 + WITH1->UU.U99.y2) * ryy) / (log_scale0 * 2L),
							((WITH1->x1 - WITH1->UU.U99.x2) * rxx +
							 (WITH1->y1 - WITH1->UU.U99.y2) * ryx) / (log_scale0 * 2L),
							((WITH1->x1 - WITH1->UU.U99.x2) * ryx +
							 (WITH1->y1 - WITH1->UU.U99.y2) * ryy) / (log_scale0 * 2L),
							savecol);
					m_color(savecol);
					break;

				case 'E':
					m_ellipse(
							xx + ((WITH1->x1 + WITH1->UU.U99.x2) * rxx +
								(WITH1->y1 + WITH1->UU.U99.y2) * ryx) / (log_scale0 * 2L),
							yy + ((WITH1->x1 + WITH1->UU.U99.x2) * rxy +
								(WITH1->y1 + WITH1->UU.U99.y2) * ryy) / (log_scale0 * 2L),
							((WITH1->x1 - WITH1->UU.U99.x2) * rxx +
							 (WITH1->y1 - WITH1->UU.U99.y2) * ryx) / (log_scale0 * 2L),
							((WITH1->x1 - WITH1->UU.U99.x2) * ryx +
							 (WITH1->y1 - WITH1->UU.U99.y2) * ryy) / (log_scale0 * 2L),
							m_curcolor());
					break;

				case 'c':
				case 'F':
					drawvector4(&WITH->vector[i], (long)(xx + (WITH1->x1 * rxx +
									WITH1->y1 * ryx) / log_scale0), (long)(yy +
									(WITH1->x1 * rxy + WITH1->y1 * ryy) / log_scale0), (long)(xx +
										(WITH1->UU.U99.x2 * rxx + WITH1->UU.U99.y2 * ryx) / log_scale0),
							(long)(yy +
								(WITH1->UU.U99.x2 * rxy + WITH1->UU.U99.y2 * ryy) / log_scale0),
							(long)(xx +
								(WITH1->UU.U99.x3 * rxx + WITH1->UU.U99.y3 * ryx) / log_scale0),
							(long)(yy +
								(WITH1->UU.U99.x3 * rxy + WITH1->UU.U99.y3 * ryy) / log_scale0),
							(long)(xx +
								(WITH1->UU.U99.x4 * rxx + WITH1->UU.U99.y4 * ryx) / log_scale0),
							(long)(yy + (WITH1->UU.U99.x4 * rxy + WITH1->UU.U99.y4 * ryy) /
								log_scale0));
					break;

				case 't':
					drawtext(&WITH->vector[i], (long)gg.scale, (long)z,
							(long)(xx + (WITH1->x1 * rxx + WITH1->y1 * ryx) / log_scale0),
							(long)(yy + (WITH1->x1 * rxy + WITH1->y1 * ryy) / log_scale0));
					break;
			}
		}
		hidecursor();
	}
	if ((c == gg.color.gate || c == gg.color.dimgate) && !suppressdots)
		drawgatedotsc(x, y, g, gg.color.gatepin);
}


void signallabel(short x, short y, short gtype, short sig, short c)
{
	log_krec *k;
	char STR1[256];

	k = kind[(gtype & (log_kindoffset - 1)) - 1];
	if (!(sig != 0 && k->flag.U3.named))
		return;
	remcursor();
	m_color((long)c);
	if (k->flag.U3.nright != (gtype / log_kindoffset == 0))
		drawstr2((int)(x - m_strwidth(NULL,
						strrtrim(strcpy(STR1, gg.signaltab[sig - 1].name))) + NAMEGAP),
				y - 3, gg.signaltab[sig - 1].name);
	else
		drawstr2(x - NAMEGAP, y - 3, gg.signaltab[sig - 1].name);
}


void gsignallabel(short x, short y, log_grec *g, short c)
{
	signallabel((int)(x * gg.scale - gg.xoff), (int)(y * gg.scale - gg.yoff),
			g->g, g->sig, c);
}


typedef enum {inside,north,east,south,west} placement;

void drawpnums(log_grec *g, short c)
{
	short rxx, rxy, ryx, ryy, i;
	char s[7];
	log_krec *WITH;
	short FORLIM;
	int tx,ty,ex,ey;
	int minx,miny,maxx,maxy;
	placement p = inside;
	long hw;

	if (zoom <= 0 || gg.textinvisible || gg.pnuminvisible)
		return;
	m_color((long)c);
	rxx = log_rotxx[g->rot];
	rxy = log_rotxy[g->rot];
	ryx = log_rotyx[g->rot];
	ryy = log_rotyy[g->rot];
	WITH = g->kind;

	/* use grab rectangle to determine which edge this pin is attached to. */
	minx = ( WITH->x1 * rxx + WITH->y1 * ryx);
	miny = ( WITH->x1 * rxy + WITH->y1 * ryy);
	maxx = ( WITH->x2 * rxx + WITH->y2 * ryx);
	maxy = ( WITH->x2 * rxy + WITH->y2 * ryy);

	FORLIM = WITH->numpnums;
	for (i = 0; i < FORLIM; i++)
	{
		sprintf(s, "%d", WITH->pnum[i].num);
		hw = strlen(s) * 5 / 2;
		ex = (WITH->pnum[i].x * rxx + WITH->pnum[i].y * ryx);
		ey = (WITH->pnum[i].x * rxy + WITH->pnum[i].y * ryy);
		if (ex < minx)
		{
			p = west;
		}
		else if (ex > maxx)
		{
			p = east;
		}
		else if (ey < miny)
		{
			p = north;
		}
		else if (ey > maxy)
		{
			p = south;
		}
		tx = g->x * gg.scale + ex * gg.scale / log_scale0 - gg.xoff;
		ty = g->y * gg.scale + ey * gg.scale / log_scale0 - gg.yoff;
		switch (p)
		{
			case west:
				tx -= 5;
				break;
			case east:
				tx += 5;
				break;
			case north:
				tx -= hw;
				ty -= 9;
				break;
			case south:
				tx -= hw;
				ty += 5;
				break;

			default:
				break;
		}

		if (p == west)
			m_rightstr(tx,ty,NULL,s);
		else
			m_drawstr(tx,ty,NULL,s);
	}
}

#undef fromtab
#undef totab


void setscale (short s);


void kdrawgatec(short i, short c)
{
	short x, y;
	short ozoom = zoom;

	x = (i - 1) * kindgroupspacing + kindgroupstart;
	y = (down + baseline) / 2;
	udrawgatec(x, y, kindgroup[i - 1], c);
	setscale(0);
	if (kindgroup[i - 1] != 0)
		signallabel(x, y, kindgroup[i - 1], kindsig[i - 1], c);
	setscale(ozoom);
}


void drawgatexc(log_grec *g, short c)
{
	if (g->dimcolor && c == gg.color.gate)
		c = gg.color.dimgate;
	drawgatec(g->x, g->y, g->g, c);
	if (!gg.textinvisible)
		gsignallabel(g->x, g->y, g, gg.color.signal);
	if (g->kind->numpnums > 0)
		drawpnums(g, gg.color.pinnum);
}

/// Draw a gate.  If it is a TO or FROM gate, draw its name as well.
void drawgatex(log_grec *g)
{
	drawgatexc(g, gg.color.gate);
}


void setdimgate(log_grec *g, int dim)
{
	log_grec *g1;

	if (g->dimcolor == dim)
		return;
	g->dimcolor = dim;
	if (gg.showpage <= 0)
		return;
	g1 = gg.pages[gg.showpage - 1]->gbase;
	while (g1 != NULL && g1 != g)
		g1 = g1->next;
	if (g1 != g)
		return;
	m_saveclip();
	clipon();
	drawgatex(g);
	m_unclip();
}

/// Draw a solder point at a certain position (grid coordinates).
void drawsolderc(short x, short y, short c)
{
	if (!showsolder)
		return;
	if (vlsi)
	{
		if (c == gg.color.conflict || c == gg.color.backgr)
			m_color((long)c);
		else
			m_color((long)gg.color.vlsicut);
		x = x * gg.scale - gg.xoff;
		y = y * gg.scale - gg.yoff;
		m_drawline(x - 3L, y - 3L, x + 3L, y + 3L);
		m_drawline(x - 3L, y + 3L, x + 3L, y - 3L);
		m_drawline(x - 2L, y - 3L, x + 4L, y + 3L);
		m_drawline(x - 2L, y + 3L, x + 4L, y - 3L);
		return;
	}
	m_color((long)c);
	x = x * gg.scale - gg.xoff;
	y = y * gg.scale - gg.yoff;
	m_drawline(x - 1L, y - 2L, x + 1L, y - 2L);
	m_drawline(x - 2L, y - 1L, x + 2L, y - 1L);
	m_drawline(x - 2L, (long)y, x + 2L, (long)y);
	m_drawline(x - 2L, y + 1L, x + 2L, y + 1L);
	m_drawline(x - 1L, y + 2L, x + 1L, y + 2L);
}


/// Draw a VLSI transistor at a certain position (grid coordinates).
void drawtransc(short x, short y, short c)
{
	if (!vlsi)
		return;
	if (c == gg.color.solder)
		m_color((long)gg.color.vlsicut);
	else
		m_color((long)c);
	x = x * gg.scale - gg.xoff;
	y = y * gg.scale - gg.yoff;
	m_circle((long)x, (long)y, 4L);
}

void drawboxc0(short x1, short y1, short x2, short y2, short c)
{
	m_color((long)c);
	m_linestyle(1L);
	rect(x1, y1, x2, y2);
	m_linestyle(0L);
}

/// Draw a dashed box on the window.
void drawboxc(log_brec *b, short  c)
{
	drawboxc0(b->x1, b->y1, b->x2, b->y2, c);
}


/// Draw the printing markers on the window.
void drawmarkersc(short c)
{
	clipon();
	m_color((long)c);
	hline(gg.markerx1, gg.markerx1 + 2, gg.markery1);
	vline(gg.markerx1, gg.markery1, gg.markery1 + 2);
	hline(gg.markerx2 - 2, gg.markerx2, gg.markery2);
	vline(gg.markerx2, gg.markery2 - 2, gg.markery2);
	clipoff();
}


/// Draw all wires and solder points in a particular node, in color C.
void drawnodec(log_nrec *n, int c)
{
	log_hwrec *hw;
	log_vwrec *vw;
	log_srec *s;

	if (gg.invisible || (gg.showconflicts == true && n->conflict == false))
		return;

	hidecursor();
	if ((unsigned)c > 15)
	{
		if (gg.glowmode && gg.showpage > 0)
			c = glowcol(n, (long)gg.color.wire[0]);
		else
			c = gg.color.wire[0];
	}
	if (vlsi && c == gg.color.wire[0])
	{
		hw = gg.pages[gg.curpage - 1]->hwbase;
		while (hw != NULL)
		{
			if (hw->node == n && hw != gg.movinghw)
			{
				m_color((long)gg.color.wire[hw->wcolr - log_wcol_normal]);
				hline(hw->x1, hw->x2, hw->y);
			}
			hw = hw->next;
		}
		vw = gg.pages[gg.curpage - 1]->vwbase;
		while (vw != NULL)
		{
			if (vw->node == n && vw != gg.movingvw)
			{
				m_color((long)gg.color.wire[vw->wcolr - log_wcol_normal]);
				vline(vw->x, vw->y1, vw->y2);
			}
			vw = vw->next;
		}
	}
	else
	{
		m_color((long)c);
		hw = gg.pages[gg.curpage - 1]->hwbase;
		while (hw != NULL) {
			if (hw->node == n && hw != gg.movinghw)
				hline(hw->x1, hw->x2, hw->y);
			hw = hw->next;
		}
		vw = gg.pages[gg.curpage - 1]->vwbase;
		while (vw != NULL)
		{
			if (vw->node == n && vw != gg.movingvw)
				vline(vw->x, vw->y1, vw->y2);
			vw = vw->next;
		}
	}
	if (showsolder)
	{
		s = gg.pages[gg.curpage - 1]->sbase;
		while (s != NULL)
		{
			if ((s->hwire != NULL && s->hwire->node == n) ||
					(s->vwire != NULL && s->vwire->node == n))
				drawsolderc(s->x, s->y, c);
			s = s->next;
		}
	}
	hidecursor();
}


/// Draw all wires and solder points in a particular node.
void drawnode(log_nrec *n)
{
	drawnodec(n, -1);
}


/// Refresh the page-number display.
void refrpagedisp()
{
	short num;

	m_color((long)gg.color.page1);
	drawstr2(across - 49, 2, "PAGE");
	m_color((long)gg.color.page2);
	drawnum2(across - 14, 2, (int)gg.curpage);
	m_color((long)gg.color.page3);
	drawstr2(across - 35, 14, "OF");
	m_color((long)gg.color.page4);
	num = gg.numpages;
	while (num > gg.curpage && isPageEmpty(gg.pages[num - 1]))
		num--;
	drawnum2(across - 14, 14, num);
}


/// Refresh the editing mode in the menus.
void refreditmode()
{
	hidecursor();
	m_color((long)gg.color.backgr);
	m_fillrect(across + menux4 - 1L, line2 - 2L, across + menux4 + 30L,
			line2 + 9L);
	switch (cureditmode)
	{

		case 1:
			m_color((long)gg.color.menuword);
			drawstr2(across + menux4, line2, "ROT");
			break;

		case 2:
			m_color((long)gg.color.menuword);
			drawstr2(across + menux4, line2, "MIR-");
			break;

		case 3:
			m_color((long)gg.color.menuword);
			drawstr2(across + menux4, line2, "MIR|");
			break;

		case 4:
			m_color((long)gg.color.selword);
			drawstr2(across + menux4, line2, "CNFG");
			break;
	}
	hidecursor();
}


/// Refresh the baseline to the correct color.
void refrbaseline(int always)
{
	short c;

	if (anyconflicts)
		c = gg.color.conflict;
	else
		c = gg.baselinecolor;
	if (!(c != gg.color.curbaseline || always))
		return;
	clipoff();
	gg.color.curbaseline = c;
	hidecursor();
	m_color((long)c);
	m_fillrect(0L, (long)baseline, (long)kindgroupleft, baseline + 1L);
	m_fillrect((long)kindgroupright, (long)baseline, (long)across,
			baseline + 1L);
	hidecursor();
}


#define rtn             "REFRESH"

/// Refresh the window (does not clear, so may be used to "clean up" existing window).
void refresh()
{
	long x1, y1, x2, y2;
	log_grec *g;
	log_hwrec *hw;
	log_vwrec *vw;
	log_nrec *n;
	log_srec *s;
	log_lrec *l;
	log_brec *b;

	stamp(&gg.refrstamp);
	gg.showpage = gg.curpage;
	showsolder = (zoom > -1 && (vlsi || !gg.glowmode || glowsolder));
	x1 = (gg.xoff - 30) / gg.scale;
	y1 = (gg.yoff - 30) / gg.scale;
	x2 = (gg.xoff + across + 30) / gg.scale;
	y2 = (gg.yoff + down + 30) / gg.scale;
	refrtimer = -1;
	remcursor();
	clipon();
	refrpagedisp();
	if (!gg.invisible && !gg.showconflicts)
	{
		suppressdots = gg.dotsvisible;
		g = gg.pages[gg.curpage - 1]->gbase;
		while (g != NULL)
		{
			if (P_ibetweenm(x1, (long)g->x, x2, (long)g->kind->bbmax) &&
					P_ibetweenm(y1, (long)g->y, y2, (long)g->kind->bbmax))
				drawgatex(g);
			g = g->next;
		}
		suppressdots = false;
		if (gg.glowmode)
		{
			hw = gg.pages[gg.curpage - 1]->hwbase;
			while (hw != NULL)
			{
				if (hw->y > y1 && hw->y < y2 && hw != gg.movinghw)
				{
					m_color(glowcol(hw->node,
								(long)gg.color.wire[hw->wcolr - log_wcol_normal]));
					hline(hw->x1, hw->x2, hw->y);
				}
				hw = hw->next;
			}
			vw = gg.pages[gg.curpage - 1]->vwbase;
			while (vw != NULL)
			{
				if (vw->x > x1 && vw->x < x2 && vw != gg.movingvw)
				{
					m_color(glowcol(vw->node,
								(long)gg.color.wire[vw->wcolr - log_wcol_normal]));
					vline(vw->x, vw->y1, vw->y2);
				}
				vw = vw->next;
			}
			if (showsolder)
			{
				s = gg.pages[gg.curpage - 1]->sbase;
				while (s != NULL)
				{
					if (s->hwire != NULL)
						n = s->hwire->node;
					else
						n = s->vwire->node;
					drawsolderc(s->x, s->y, (int)glowcol(n, (long)gg.color.solder));
					s = s->next;
				}
			}
		}
		else
		{
			m_color((long)gg.color.wire[0]);
			hw = gg.pages[gg.curpage - 1]->hwbase;
			while (hw != NULL)
			{
				if (hw->y > y1 && hw->y < y2 && hw != gg.movinghw)
				{
					if (vlsi)
						m_color((long)gg.color.wire[hw->wcolr - log_wcol_normal]);
					hline(hw->x1, hw->x2, hw->y);
				}
				hw = hw->next;
			}
			vw = gg.pages[gg.curpage - 1]->vwbase;
			while (vw != NULL)
			{
				if (vw->x > x1 && vw->x < x2 && vw != gg.movingvw)
				{
					if (vlsi)
						m_color((long)gg.color.wire[vw->wcolr - log_wcol_normal]);
					vline(vw->x, vw->y1, vw->y2);
				}
				vw = vw->next;
			}
			if (showsolder)
			{
				s = gg.pages[gg.curpage - 1]->sbase;
				while (s != NULL)
				{
					if (s->hwire != NULL)
						n = s->hwire->node;
					else
						n = s->vwire->node;
					drawsolderc(s->x, s->y, gg.color.solder);
					s = s->next;
				}
			}
		}

		if (gg.dotsvisible)
		{
			g = gg.pages[gg.curpage - 1]->gbase;
			while (g != NULL)
			{
				if (g->x > x1 && g->x < x2 && g->y > y1 && g->y < y2)
					drawgatedotsc(g->x, g->y, g->g, gg.color.gatepin);
				g = g->next;
			}
		}
	}
	else
	{
		g = gg.pages[gg.curpage - 1]->gbase;
		while (g != NULL)
		{
			if (g->kind->flag.U3.visible)
				drawgatex(g);
			g = g->next;
		}
	}
	if (!gg.textinvisible)
	{
		b = gg.pages[gg.curpage - 1]->bbase;
		while (b != NULL)
		{
			drawboxc(b, gg.color.dashbox);
			b = b->next;
		}
		l = gg.pages[gg.curpage - 1]->lbase;
		m_color((long)gg.color.labeltext);
		while (l != NULL)
		{
			if (l->y > y1 && l->y < y2)
			{
				m_drawstr(l->x * gg.scale - gg.xoff, l->y * gg.scale - gg.yoff + 2,
						logfont_lfont, l->name);
			}
			l = l->next;
		}
	}
	if (gg.markers)
		drawmarkersc(gg.color.marker);
	clipoff();
	restorecursor();
}

#undef rtn

/// Window is messed up, so refresh it as soon as the user stops moving around.
void refreshsoon()
{
	refrtimer0 = refrdelay;
	refrtimer = refrdelay;
	gg.refrflag = true;
}

/// Clear and redraw menu area of window.
void refrmenu()
{
	short i, FORLIM;

	remcursor();
	uerase(0, baseline + 2, across, down);
	clipoff();
	refrbaseline(true);
	m_color((long)gg.color.backgr);
	m_fillrect(kindgroupleft + 1L, (long)baseline, kindgroupright - 1L,
			baseline + 1L);
	m_color((long)gg.color.menuword);
	drawstr2(menux1, line1, "Frills");
	drawstr2(menux2, line1, "Editing");
	drawstr2(menux1, line2, "HELP");
	drawstr2(menux2, line2, "CAT");
	drawstr2(across + menux3, line1, "Cursor");
	drawstr2(across + menux4, line1, "Misc");
	refreditmode();
	if (modeflag)
		log_setmode(modename);
	else
		log_setmode("");
	FORLIM = kindgroupsize;
	for (i = 1; i <= FORLIM; i++)
		kdrawgatec(i, gg.color.kindgate);
}


void beginbottom()
{
	remcursor();
	clipoff();
	bottomcount++;
	if (m_autoraise)
	{
		clearshowalpha();
	}
	else
	{
		clearalpha();
	}
	nc_gotoXY(0, txdown - 1);
}


void endbottom()
{
	remcursor();
	bottomcount--;
	if (m_autoraise)
	{
		if (bottomcount < 1)   /* fix from WES */
		{
			bottomcount =0;
			m_graphics_on();
		}
	}
	else
	{
		m_graphics_on();   /* switch back to the graphics window */
	}
	clearalpha();
}

/// Clear and redraw working area of window.
void refrwindow1()
{
	m_colormode((long)m_normal);
	m_noclip();
	remcursor();
	uerase(0, 0, across, baseline - 1);
	resetmessages();
	refresh();
	gg.refrflag = true;
}


/// Clear and redraw entire window.
void refrwindow()
{
	gg.showpage = gg.curpage;
	bottomcount = 0;
	m_colormode((long)m_normal);
	m_noclip();
	fixcolormap();
	clearalpha();
	refrmenu();
	refrwindow1();
	m_graphics_on();
}

void resize_window()
{
	initwindow2();
	initwindow();
}

//// Switch window to catalog (TRUE) or circuit (FALSE).
void showcatalog()
{
	short i;
	catboxrec *cb;
	short x1, y1, x2, y2, FORLIM;

	clearwindow();
	gg.showpage = log_page_cat;
	FORLIM = maxkinds;
	for (i = 1; i <= FORLIM; i++)
	{
		if (kind[i - 1] != NULL)
		{
			udrawgatec((int)(((i - 1) % catwidth * 2 + 1) * gridcen),
					(int)(((i - 1) / catwidth * 2 + 1) * gridcen), i,
					gg.color.catgate);
		}
	}
	cb = catboxes;
	while (cb != NULL)
	{
		x1 = (cb->pos - 1) % catwidth * gridcen * 2;
		y1 = (cb->pos - 1) / catwidth * gridcen * 2;
		x2 = ((cb->pos + cb->count - 2) % catwidth + 1) * gridcen * 2;
		y2 = ((cb->pos + cb->count - 2) / catwidth + 1) * gridcen * 2;
		m_color((long)catboxcolor);
		m_linestyle(2L);
		m_drawline((long)x1, (long)y1, (long)x2, (long)y1);
		m_linestyle(2L);
		m_drawline((long)x1, (long)y2, (long)x2, (long)y2);
		m_linestyle(2L);
		m_drawline((long)x1, (long)y1, (long)x1, (long)y2);
		m_linestyle(2L);
		m_drawline((long)x2, (long)y1, (long)x2, (long)y2);
		m_linestyle(0L);
		cb = cb->next;
	}
}

void init_X_window()
{
	setup_log_cursors();
	choose_log_cursor(0);
}

short getcursorxposition()
{
	return cursx;
}

short getcursoryposition()
{
	return cursy;
}

int getcursorflag()
{
	return cursorflag;
}

